// $Id: PowspecFiltbank.cc,v 1.3 1999/01/01 01:11:51 dpwe Exp dpwe $
// 
// $Log: PowspecFiltbank.cc,v $
// Revision 1.3  1999/01/01 01:11:51  dpwe
// *** empty log message ***
//
// Revision 1.2  1998/06/08 18:13:31  bedk
// Added Bark-scale filters
//
// Revision 1.1  1998/05/28 21:12:34  bedk
// Initial revision
//
// 
// Simulates a filterbank by generating FFT power spectra and doing a
// weighted accumulation of the power spectral points.
// 

#include <assert.h>
#include <math.h>

#include "PowspecFiltbank.h"

/* 1998dec31 dpwe@icsi.berkeley.edu
 * I've added a new optional arg, nyq_frac, which is the fraction of 
 * the total bandwidth to use in calculating the number of channels 
 * and their widths.  Thus nyq_frac = 0.5 treats 16 kHz audio as 
 * if it had been downsampled to 8 kHz.  nyq_frac=1.0 (the default) 
 * performs as before. 
 */

PowspecFiltbank::PowspecFiltbank(const size_t a_IOBlocking,
                                 const size_t a_nfft,
                                 const size_t a_winStep,
                                 const Vector& a_window,
                                 const float a_bw,
                                 const float a_filterStep,
                                 const float a_tw,
                                 const float a_sf,
                                 const float a_loCutoff,
                                 const FilterShape a_filterShape,
                                 const FilterScale a_filterScale,
                                 const bool a_dither, 
				 const float a_nyq_frac)
  : IOBlocking(a_IOBlocking),
    nfft(a_nfft),
    winStep(a_winStep),
    indexMax(1 + (nfft >> 1)),
    window(a_window),
    bufLen(0),
    pointsToNextWindow(0),
    inBuffer(new float[window.m - 1]),
    FFTBuffer(new float[nfft << 1]),
    win_endp(FFTBuffer + (window.m << 1)),
    fft_endp(FFTBuffer + (nfft << 1)),
    reorder(new float[1 + (nfft >> 1)]),
    basis(new float[nfft]),
    index(new size_t[nfft]),
    bw(a_bw),
    filterStep(a_filterStep),
    tw(a_tw),
    sf(a_sf),
    loCutoff(a_loCutoff),
    filterShape(a_filterShape),
    filterScale(a_filterScale),
    dither(a_dither),
    nyq_frac(a_nyq_frac), 
    numChannels(PowspecFiltbank::computeNumChannels(filterScale,
                                                    bw,
                                                    filterStep,
                                                    sf,
                                                    loCutoff, 
						    nyq_frac)),
    outBuffer(new float[numChannels * (1 + (size_t) (IOBlocking/winStep))])
{
  assert(isPowerOf2(nfft));
  assert(nfft >= window.m);
  assert(window.m > 1);
  assert(IOBlocking > 0);
  assert(winStep > 0);

  // precompute basis for FFT
  // 
  initBasis(basis, nfft);
  
  // precompute bit-reversed index offsets into output buffer
  // 
  size_t* ip = index;
  size_t bitRev = 0;
  for (size_t k = 0; k < (nfft << 1); k += 2)
    {
      *ip++ = bitRev;
      for (size_t m = (nfft >> 1); m != 0; m >>= 1)
        {
          if (bitRev & m)
            {
              bitRev &= ~m;
            }
          else
            {
              bitRev |= m;
              break;
            }
        }
    }

  // Compute the starting points for and number of points in each
  // filter
  //
  filt_npts = new size_t[numChannels];
  filt_start = new size_t[numChannels];

  size_t tot_pts = 0;

  switch (filterShape)
    {
    case TRAPEZOIDAL:
      switch (filterScale)
        {
        case CONST_Q:
          {
            for (size_t i = 0; i < numChannels - 1; i++)
              {
                size_t k = numChannels - i - 1;
                size_t hi_bin = 
                  (size_t)(nfft * pow(2.0, -1.0 - filterStep * k + tw));
                size_t lo_bin = 
                  (size_t)(1.0 + nfft *
                           pow(2.0, -1.0 - filterStep * k - bw - tw));
                size_t n = hi_bin - lo_bin + 1;
                filt_npts[i] = n;
                tot_pts += n;
                filt_start[i] = lo_bin;
              }
            size_t hi_bin = (nfft >> 1);
            size_t lo_bin = (size_t)(1.0 + nfft * pow(2.0, -1.0 - bw - tw));
            size_t n = hi_bin - lo_bin + 1;
            filt_npts[numChannels - 1] = n;
            tot_pts += n;
            filt_start[numChannels - 1] = lo_bin;
          }
        break;

        case BARK:
          {
            // Nyquist frequency in Barks
            //
            const double nyq_b = 6.0 * asinh(sf / 1200.0);
            
            for (size_t i = 0; i < numChannels - 1; i++)
              {
                size_t k = numChannels - i - 1;

                size_t hi_bin = 
                  (size_t)(nfft * 600.0 * 
                           sinh((nyq_b - filterStep * k + tw) / 6.0) /
                           sf); 

                size_t lo_bin = 
                  (size_t)(1.0 + nfft * 600.0 * 
                           sinh((nyq_b - filterStep * k - bw - tw) / 
                                6.0) /
                           sf);

                size_t n = hi_bin - lo_bin + 1;
                filt_npts[i] = n;
                tot_pts += n;
                filt_start[i] = lo_bin;
              }
            size_t hi_bin = (nfft >> 1);
            size_t lo_bin = 
              (size_t)(1.0 + nfft * 600.0 * 
                       sinh((nyq_b - bw - tw) / 6.0) / sf);
            size_t n = hi_bin - lo_bin + 1;
            filt_npts[numChannels - 1] = n;
            tot_pts += n;
            filt_start[numChannels - 1] = lo_bin;
          }
        break;
        }
      break;

    case TRIANGULAR:
      switch (filterScale)
        {
        case CONST_Q:
          {
            for (size_t i = 0; i < numChannels - 1; i++)
              {
                size_t k = numChannels - i - 1;
                size_t hi_bin =
                  (size_t)(nfft * 
                           pow(2.0, -1.0 - filterStep * k));
                size_t lo_bin = 
                  (size_t)(1.0 + nfft *
                           pow(2.0, -1.0 - filterStep * k - 2.0 * bw));
                size_t n = hi_bin - lo_bin + 1;
                filt_npts[i] = n;
                tot_pts += n;
                filt_start[i] = lo_bin;
              }
            size_t hi_bin = (nfft >> 1);
            size_t lo_bin = 
              (size_t)(1.0 + nfft * pow(2.0, -1.0 - 2.0 * bw));
            size_t n = hi_bin - lo_bin + 1;
            filt_npts[numChannels - 1] = n;
            tot_pts += n;
            filt_start[numChannels - 1] = lo_bin;
          }
        break;
        
        case BARK:
          {
	    // Number of channels if we weren't fractionated with nyq_frac
	    const int expectNumChans = 1.0 + ceil((6.0 * (asinh(sf/1200.0) - 
						       asinh(loCutoff/600.0)) -
						   bw) /
						  filterStep);

            // Nyquist frequency in Barks
            //
            const double nyq_b = 6.0 * asinh(sf / 1200.0);
            
            for (size_t i = 0; i < numChannels - 1; i++)
              {
                size_t k = expectNumChans - i - 1;
                size_t hi_bin =
                  (size_t)(nfft * 600.0 *
                           sinh((nyq_b - filterStep * k) / 6.0) / sf);
                size_t lo_bin =
                  (size_t)(1.0 + nfft * 600.0 *
                           sinh((nyq_b - filterStep * k - 2.0 * bw) / 6.0) /
                           sf);
                size_t n = hi_bin - lo_bin + 1;
                filt_npts[i] = n;
                tot_pts += n;
                filt_start[i] = lo_bin;
		//fprintf(stderr, "chan %d lo=%d hi=%d\n", i, lo_bin, hi_bin);
              }
            size_t hi_bin = (int)(0.5 + nyq_frac*(nfft >> 1));
            size_t lo_bin =
              (size_t)(1.0 + 
                       hi_bin * 1200.0 * sinh((nyq_b - 2.0 * bw) / 6.0) / sf);
            size_t n = hi_bin - lo_bin + 1;
            filt_npts[numChannels - 1] = n;
            tot_pts += n;
            filt_start[numChannels - 1] = lo_bin;
	    //fprintf(stderr, "chan %d lo=%d hi=%d\n", numChannels-1, lo_bin, hi_bin);
          }
        break;
        }
    }

  // Compute the filter weights
  //
  filt_weights = new float[tot_pts];
  float* wp = filt_weights;

  switch (filterShape)
    {
    case TRAPEZOIDAL:
      switch (filterScale)
        {
        case CONST_Q:
          {
            size_t i = 0;
            for (;
                 i < numChannels - 1;
                 i++)
              {
                size_t j = filt_start[i];
                size_t k = numChannels - i - 1;
                
                // Find boundaries of transition regions
                //
                double hi_tr_f = 
                  sf * pow(2.0, -1.0 - filterStep * k);
                double lo_tr_f =
                  sf * pow(2.0, -1.0 - filterStep * k - bw);
                size_t hi_tr_b = (size_t)(1 + nfft * hi_tr_f / sf);
                size_t lo_tr_b = (size_t)(nfft * lo_tr_f / sf);
                
                // Calculate FFT bin weights in low transition region
                //
                for (;
                     j <= lo_tr_b;
                     j++)
                  {
                    *wp++ = 1.0 + log((j * sf) / (nfft * lo_tr_f)) /
                      (tw * log(2.0));
                    // float w = 1.0 + log((j * sf) / (nfft * lo_tr_f)) /
                    // (tw * log(2.0));
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
                
                // FFT bin weights are unity in passband
                //
                for (;
                     j < hi_tr_b;
                     j++)
                  {
                    *wp++ = 1.0;
                    // float w = 1.0;
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
                
                // Calculate FFT bin weights in high transition region
                //
                for (;
                     j < filt_start[i] + filt_npts[i];
                     j++)
                  {
                    *wp++ = 1.0 - log((j * sf) / (nfft * hi_tr_f)) /
                      (tw * log(2.0));
                    // float w = 1.0 - log((j * sf) / (nfft * hi_tr_f)) /
                    //   (tw * log(2.0));
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
              }
            
            // Find boundaries of transition regions
            //
            double lo_tr_f = sf * pow(2.0, -1.0 - bw);
            size_t lo_tr_b = (size_t)(nfft * lo_tr_f / sf);
            
            
            // Calculate FFT bin weights in low transition region
            //
            size_t j = filt_start[i];
            
            for (;
                 j <= lo_tr_b;
                 j++)
              {
                *wp++ = 1.0 + log((j * sf) / (nfft * lo_tr_f)) /
                  (tw * log(2.0));
                // float w = 1.0 + log((j * sf) / (nfft * lo_tr_f)) /
                //  (tw * log(2.0));
                // fprintf(stderr, "%d %g\n", j, w);
                // *wp++ = w;
              }
            
            // FFT bin weights are unity in passband
            //
            for (;
                 j < filt_start[i] + filt_npts[i];
                 j++)
              {
                *wp++ = 1.0;
                // float w = 1.0;
                // fprintf(stderr, "%d %g\n", j, w);
                // *wp++ = w;
              }
          }
        break;

        case BARK:
          {
            const double nyq_b = 6.0 * asinh(sf / 1200.0);
            
            size_t i = 0;
            for (;
                 i < numChannels - 1;
                 i++)
              {
                size_t j = filt_start[i];
                size_t k = numChannels - i - 1;
                
                // Find boundaries of transition regions
                //
                double hi_tr_b = nyq_b - filterStep * k;

                double lo_tr_b = nyq_b - filterStep * k - bw;

                double hi_tr_f = 600.0 * sinh(hi_tr_b / 6.0);

                double lo_tr_f = 600.0 * sinh(lo_tr_b / 6.0);

                size_t hi_tr_bin = (size_t)(1.0 + nfft * hi_tr_f / sf);

                size_t lo_tr_bin = (size_t)(nfft * lo_tr_f / sf);
                
                // Calculate FFT bin weights in low transition region
                //
                for (;
                     j <= lo_tr_bin;
                     j++)
                  {
                    *wp++ = 1.0 - 
                      (lo_tr_b - 6.0 * asinh((j * sf) / 
                                             (600.0 * nfft))) / tw;
                    // float w = 1.0 - 
                    //   (lo_tr_b - 6.0 * asinh((j * sf) / 
                    //                          (600.0 * nfft))) / tw;
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
                
                // FFT bin weights are unity in passband
                //
                for (;
                     j < hi_tr_bin;
                     j++)
                  {
                    *wp++ = 1.0;
                    // float w = 1.0;
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
                
                // Calculate FFT bin weights in high transition region
                //
                for (;
                     j < filt_start[i] + filt_npts[i];
                     j++)
                  {
                    *wp++ = 1.0 +
                      (hi_tr_b - 6.0 * asinh((j * sf) / 
                                             (600.0 * nfft))) / tw;
                    // float w = 1.0 +
                    //   (hi_tr_b - 6.0 * asinh((j * sf) / 
                    //                          (600.0 * nfft))) / tw;
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
              }
            
            // Find boundaries of transition regions
            //
            double lo_tr_b = nyq_b - bw;

            double lo_tr_f = 600.0 * sinh(lo_tr_b / 6.0);

            size_t lo_tr_bin = (size_t)(nfft * lo_tr_f / sf);
            
            // Calculate FFT bin weights in low transition region
            //
            size_t j = filt_start[i];
            
            for (;
                 j <= lo_tr_bin;
                 j++)
              {
                *wp++ = 1.0 - 
                  (lo_tr_b - 6.0 * asinh((j * sf) / 
                                         (600.0 * nfft))) / tw;
                // float w = 1.0 - 
                //   (lo_tr_b - 6.0 * asinh((j * sf) / 
                //                          (600.0 * nfft))) / tw;
                // fprintf(stderr, "%d %g\n", j, w);
                // *wp++ = w;
              }
            
            // FFT bin weights are unity in passband
            //
            for (;
                 j < filt_start[i] + filt_npts[i];
                 j++)
              {
                *wp++ = 1.0;
                // float w = 1.0;
                // fprintf(stderr, "%d %g\n", j, w);
                // *wp++ = w;
              }
          }
        break;
        }
      break;

    case TRIANGULAR:
      switch (filterScale)
        {
        case CONST_Q:
          {
            size_t i = 0;
            for (;
                 i < numChannels;
                 i++)
              {
                size_t j = filt_start[i];
                size_t k = (numChannels - 1) - i;
                
                // Find filter lower cutoff, upper cutoff, and center
                // frequency
                //
                double lo_f = 
                  sf * pow(2.0, -1.0 - filterStep * k - 2.0 * bw);
                
                double c_f = 
                  sf * pow(2.0, -1.0 - filterStep * k - bw);
                
                double hi_f = sf * pow(2.0, -1.0 - filterStep * k);
                
                size_t c_b = (size_t)(nfft * c_f / sf);
                
                // Compute weights for bins less than center frequency
                //
                for (;
                     j <= c_b;
                     j++)
                  {
                    *wp++ =
                      ((double)(j * sf / nfft) - lo_f) / (c_f - lo_f);
                    // float w =
                    //   ((double)(j * sf / nfft) - lo_f) / (c_f - lo_f);
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
                
                // Compute weights for bins greater than center
                // frequency
                //
                for (;
                     j < filt_start[i] + filt_npts[i];
                     j++)
                  {
                    *wp++ =
                      (hi_f - (double)(j * sf / nfft)) / (hi_f - c_f);
                    // float w =
                    //   (hi_f - (double)(j * sf / nfft)) / (hi_f - c_f);
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
              }
          }
        break;

        case BARK:
          {
	    // Number of channels if we weren't fractionated with nyq_frac
	    const int expectNumChans = 1.0 + ceil((6.0 * (asinh(sf/1200.0) - 
						       asinh(loCutoff/600.0)) -
						   bw) /
						  filterStep);

            const double nyq_b = 6.0 * asinh(sf / 1200.0);
            
            size_t i = 0;
            for (;
                 i < numChannels;
                 i++)
              {
                size_t j = filt_start[i];
                size_t k = (expectNumChans - 1) - i;
                
                // Find filter lower cutoff, upper cutoff, and center
                // frequency
                //
                double lo_f = 600.0 *
                  sinh((nyq_b - filterStep * k - 2.0 * bw) / 6.0);
                
                double c_f = 600.0 *
                  sinh((nyq_b - filterStep * k -  bw) / 6.0);
                
                double hi_f = 600.0 *
                  sinh((nyq_b - filterStep * k) / 6.0);
                
                size_t c_b = (size_t)(nfft * c_f / sf);
                
		// ** Scale all weights by nyq_frac to compensate for 
		//    (presumed) longer FFT window

                // Compute weights for bins less than center frequency
                //
                for (;
                     j <= c_b;
                     j++)
                  {
                    *wp++ = nyq_frac * nyq_frac *
                      ((double)(j * sf / nfft) - lo_f) / (c_f - lo_f);
                    // float w =
                    //   ((double)(j * sf / nfft) - lo_f) / (c_f - lo_f);
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
                
                // Compute weights for bins greater than center
                // frequency
                //
                for (;
                     j < filt_start[i] + filt_npts[i];
                     j++)
                  {
                    *wp++ = nyq_frac * nyq_frac *
                      (hi_f - (double)(j * sf / nfft)) / (hi_f - c_f);
                    // float w =
                    //   (hi_f - (double)(j * sf / nfft)) / (hi_f - c_f);
                    // fprintf(stderr, "%d %g\n", j, w);
                    // *wp++ = w;
                  }
              }
          }
        break;
        }
      break;
    }
}

PowspecFiltbank::~PowspecFiltbank()
{
  delete [] FFTBuffer;
  delete [] inBuffer;
  delete [] basis;
  delete [] index;
  delete [] filt_weights;
  delete [] filt_start;
  delete [] filt_npts;
  delete [] outBuffer;
  delete [] reorder;
}

// Compute the power spectrum of an input signal using the short-time
// Fourier transform
// 
const CMatrix
PowspecFiltbank::processBlock(const CVector in,
                              bool& lastTime)
{
  assert(in.m <= IOBlocking);

  if (state == Start) { state = Continuing; }

  if (bufLen + in.m < pointsToNextWindow + window.m)
    {
      if (lastTime)
        {
          state = End;
          return CMatrix(numChannels,
                         0, 
                         (1 + (size_t) (IOBlocking/winStep)),
                         outBuffer);
        }
      else
        {
          if (in.m <= pointsToNextWindow)
            // Chew up input
            // 
            {
              pointsToNextWindow -= in.m;
              return CMatrix(numChannels,
                             0, 
                             (1 + (size_t) (IOBlocking/winStep)),
                             outBuffer);
            }
          else
            // Copy new data into input buffer and return
            // 
            {
              const float* sp = in.data + pointsToNextWindow;
              const float* const in_endp = in.data + in.m;
              float* bp = inBuffer + bufLen;
              bufLen += in.m - pointsToNextWindow;
              pointsToNextWindow = 0;
              while (sp != in_endp) { *bp++ = *sp++; }
              return CMatrix(numChannels,
                             0, 
                             (1 + (size_t) (IOBlocking/winStep)),
                             outBuffer);
            }
        }
    }
  else
    {
      // Generate output
      // 
      const size_t nout = 1 + ((bufLen + in.m - pointsToNextWindow -
                                window.m) / winStep);
      const float* const lastPt = in.data + (nout - 1) * winStep +
        window.m - bufLen - pointsToNextWindow;
      float* op = outBuffer;
      bool straddle = (bufLen != 0);
      const float* winStart =
        straddle ? inBuffer : in.data + pointsToNextWindow; 
      const float* sp;
      const float* const buf_endp = inBuffer + bufLen;
      const float* const in_endp = in.data + in.m;
      do
        {
          sp = winStart;
          float* bp = FFTBuffer;

          // Copy data into FFT buffer
          // 
          if (straddle)
            // Data window straddles the buffer and the input
            // 
            {
              while (sp != buf_endp)
                {
                  *bp++ = *sp++;
                  *bp++ = 0.0f;
                }
              sp = in.data;
              while (bp != win_endp)
                {
                  *bp++ = *sp++;
                  *bp++ = 0.0f;
                }
              if (buf_endp > winStart + winStep)
                {
                  winStart += winStep;
                }
              else
                {
                  straddle = false;
                  winStart = in.data + winStep - (buf_endp - winStart);
                }
            }
          else
            // Data window is in the input only
            // 
            {
              while (bp != win_endp)
                {
                  *bp++ = *sp++;
                  *bp++ = 0.0f;
                }
              winStart += winStep;
            }

          // Window data
          // 
          const float* wp = window.data;
          bp = FFTBuffer;
          while (bp != win_endp)
            {
              *bp *= *wp;
              wp++;
              bp += 2;
            }
          // Zero-pad data
          // 
          while (bp != fft_endp)
            {
              *bp++ = 0.0f;
            };

          // Compute FFT of windowed data
          // 
          fft2f(FFTBuffer, nfft, basis);

          // compute power spectrum into reorder buffer
          // 
          bp = FFTBuffer;
          float* rp = reorder;
          const size_t* ip = index;
          while (bp != fft_endp)
            {
              size_t i = *ip++;
              if (i < indexMax)
                {
                  const float tr = *bp;
                  const float ti = *(bp + 1);
                  float tm = tr * tr + ti * ti;
                  if (dither) { tm += nfft; }
                  *(rp + i) = tm;
                }
              bp += 2;
            }

          // rp = reorder;
          // for (size_t i = 0; i <= (nfft >> 1); i++)
          //   {
          //     fprintf(stderr, "%g ", *(rp + i));
          //   }
          // fprintf(stderr, "\n");

          // Do dot products to compute filter outputs, write into the
          // output buffer
          //
          float* cp = op;
          float* fp = filt_weights;
          for (size_t i = 0; i < numChannels; i++)
            {
              *cp = dotproduct(filt_npts[i],
                               reorder + filt_start[i],
                               fp);
              // fprintf(stderr, "%g ", *cp);
              fp += filt_npts[i];
              cp += (1 + (size_t) (IOBlocking/winStep));
            }
          // fprintf(stderr, "\n");

          bufLen = 0;
          op++;
        }
      while (sp != lastPt);

      // If necessary, fill input buffer or set number of points to
      // the next window
      // 
      if (straddle)
        {
        // Data window straddles the buffer and the input
        // 
          sp = winStart;
          float *bp = inBuffer;
          while (sp != buf_endp) { *bp++ = *sp++; }
          sp = in.data;
          while (sp != in_endp) { *bp++ = *sp++; }
          bufLen = (buf_endp - winStart) + (in_endp - in.data);
          pointsToNextWindow = 0;
        }
      else
        // Data window is only in the input
        // 
        {
          if (winStart > in_endp)
            // Need to see some more points before we get to a new
            // window
            // 
            {
              bufLen = 0;
              pointsToNextWindow = (winStart - in_endp);
            }
          else
            // Have data left over --- copy into buffer
            // 
            {
              sp = winStart;
              float *bp = inBuffer;
              while (sp != in_endp) { *bp++ = *sp++; }
              bufLen = (in_endp - winStart);
              pointsToNextWindow = 0;
            }
        }

      return CMatrix(numChannels,
                     nout,
                     (1 + (size_t) (IOBlocking/winStep)),
                     outBuffer);
    }
}

// PHIPAC code for dot-product.
//
float
PowspecFiltbank::dotproduct(const size_t n,
                            const float* avec,
                            const float* bvec)
{
  register float sum0 = 0.0;
  register float sum1 = 0.0;

  const float *all_endp = avec + n;
  const float *smallblock_endp = avec + (n & ~3);
  const float *bigblock_endp = avec + (n & ~15);

  while (avec != bigblock_endp)
    {
      sum0 += avec[0] * bvec[0];
      sum1 += avec[1] * bvec[1];
      sum0 += avec[2] * bvec[2];
      sum1 += avec[3] * bvec[3];
      sum0 += avec[4] * bvec[4];
      sum1 += avec[5] * bvec[5];
      sum0 += avec[6] * bvec[6];
      sum1 += avec[7] * bvec[7];
      sum0 += avec[8] * bvec[8];
      sum1 += avec[9] * bvec[9];
      sum0 += avec[10] * bvec[10];
      sum1 += avec[11] * bvec[11];
      sum0 += avec[12] * bvec[12];
      sum1 += avec[13] * bvec[13];
      sum0 += avec[14] * bvec[14];
      sum1 += avec[15] * bvec[15];

      avec += 16; bvec += 16;
    }

  while (avec != smallblock_endp)
    {
      sum0 += avec[0] * bvec[0];
      sum1 += avec[1] * bvec[1];
      sum0 += avec[2] * bvec[2];
      sum1 += avec[3] * bvec[3];

      avec += 4; bvec += 4;
    }

  sum0 += sum1;

  /* Do trailing portion in naive loop. */
  while (avec != all_endp)
    sum0 += (*avec++) * (*bvec++);

  return sum0;
}




#ifdef TEST

#include <math.h>

#include "SphereInput.h"

int
main(int argc, char** argv)
{
  if (argc != 5)
    {
      fprintf(stderr, "Usage: %s nfft nwin nstep nblk\n", *argv);
      exit(1);
    }

  const size_t nfft  = (size_t) atof(*(argv + 1));
  const size_t nwin  = (size_t) atof(*(argv + 2));
  const size_t nstep = (size_t) atof(*(argv + 3));
  const size_t nblk  = (size_t) atof(*(argv + 4));

  // Generate a Hamming window
  // 
  Vector hamming(nwin);
  for (size_t i = 0; i < nwin; i++)
    hamming.data[i] = 0.54 - 0.46 * cos(2.0 * M_PI * i / (nwin - 1));

  SphereInput spi(nblk, 8000, "/u/drspeech/data/numbers95/wavfile/cs/0/NU-19.zipcode.wav");
  // SphereInput spi(nblk, 100, "tstFFTPwrSpec.wav");
  PowspecFiltbank pspec(nblk, nfft, nstep, hamming, 0.25, 0.25, 0.0,
                        8000.0, 300.0, PowspecFiltbank::TRIANGULAR,
                        PowspecFiltbank::CONST_Q); 
  bool lastTime = false;

  while (!lastTime)
    {
      const CVector in = spi.processBlock(lastTime);
      const CMatrix out = pspec.processBlock(in, lastTime);
      for (size_t i = 0; i < out.n; i++)
        {
          for (size_t j = 0; j < out.m; j++)
            {
              fprintf(stdout, "%g ", out.data[i + j * out.s]);
            }
          fprintf(stdout, "\n");
        }
    }

  return 0;
}

#endif // TEST
