/*
   $Header: /u/spert/src/intvec/testsuite/RCS/vecele_vi32_vi32_test.c,v 1.3 1996/01/05 00:15:34 davidj Exp $

   Test element wise vector routines of the form op_vi32_vi32.
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"


typedef void (*rtn_t)(size_t len, const intv_int32_t* in,
		      intv_int32_t* res);

/* Performs one test given input values and expected output. */

void
onetest_normal(rtn_t rtn, size_t len,
	const intv_int32_t* in, const intv_int32_t* correct)
{
    /* Create padded copies of inputs. */
    intv_int32_t* c_in = rtst_padvec_clone_vi32(len, in);
    intv_int32_t* c_res = rtst_padvec_new_vi32(len);

    /* Run routine. */
    rtn(len, c_in, c_res);

    /* Check result. */
    rtst_checkeq_vi32vi32(len, c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_vi32vi32(len, in, c_in);

    /* Free and check padding. */
    rtst_padvec_del_vi32(c_in);
    rtst_padvec_del_vi32(c_res);
}

void
copy_6(rtn_t rtn)
{
    const intv_int32_t in[] = { 0, -1, 1, -123456, 0x7fffffff, INTV_INT32_MIN};
    const intv_int32_t correct[] = { 0, -1, 1, -123456, 0x7fffffff,INTV_INT32_MIN};
    
    onetest_normal(rtn, 6, in, correct);
}

void
neg_6(rtn_t rtn)
{
    const intv_int32_t in[] = { 0, -1, 1, -123456, 0x7fffffff, INTV_INT32_MIN};
    const intv_int32_t correct[] = { 0, 1, -1, 123456, 0x80000001, INTV_INT32_MIN};
    
    onetest_normal(rtn, 6, in, correct);
}

void
abs_6(rtn_t rtn)
{
    const intv_int32_t in[] = { 0, -1, 1, -123456, 0x7fffffff, INTV_INT32_MIN};
    const intv_int32_t correct[] = { 0, 1, 1, 123456, 0x7fffffff, INTV_INT32_MIN};

    onetest_normal(rtn, 6, in, correct);
}

void
satneg_6(rtn_t rtn)
{
    const intv_int32_t in[] = { 0, -1, 1, -123456, 0x7fffffff, INTV_INT32_MIN};
    const intv_int32_t correct[] = { 0, 1, -1, 123456, 0x80000001,
				     INTV_INT32_MAX};
    
    onetest_normal(rtn, 6, in, correct);
}

void
satabs_6(rtn_t rtn)
{
    const intv_int32_t in[] = { 0, -1, 1, -123456, 0x7fffffff, INTV_INT32_MIN};
    const intv_int32_t correct[] = { 0, 1, 1, 123456, 0x7fffffff,
				     INTV_INT32_MAX};

    onetest_normal(rtn, 6, in, correct);
}

void
countlz_6(rtn_t rtn)
{
    const intv_int32_t in[] = { 0, 1, 2, 0x3fffffff, 0x80000000, 0x01010101 };
    const intv_int32_t correct[] = { 32, 31, 30, 2, 0, 7 };

    onetest_normal(rtn, 6, in, correct);
}

void
swapb_explicit(rtn_t rtn)
{
    const intv_int32_t in[] =
    {
	0, 0x12345678, 0xfedcba98
    };
    const intv_int32_t correct[] =
    {
	0, 0x78563412, 0x98badcfe
    };

    onetest_normal(rtn, sizeof(in)/sizeof(in[0]), in, correct);
}

/* Compare the results of two vector routines */

void
random_test(rtn_t nv_rtn, rtn_t test_rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    intv_int32_t *in = rtst_padvec_new_vi32(maxlen);
    intv_int32_t *res = rtst_padvec_new_vi32(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_badrand_vi32(len, in);
        nv_rtn(len, in, res);

        rtst_log("len=%d\n", len);

        onetest_normal(test_rtn, len, in, res);
    }

    rtst_padvec_del_vi32(in);
    rtst_padvec_del_vi32(res);
}

void
test_copy(void)
{
    rtst_start("copy_vi32_vi32");
    rtst_log("testing nv.\n");
    copy_6(nv_copy_vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    copy_6(tor_copy_vi32_vi32);
    random_test(nv_copy_vi32_vi32, tor_copy_vi32_vi32);
#endif
    rtst_passed();
}

void
test_neg(void)
{
    rtst_start("neg_vi32_vi32");
    rtst_log("testing nv.\n");
    neg_6(nv_neg_vi32_vi32);
/*
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    neg_6(tor_neg_vi32_vi32);
    random_test(nv_neg_vi32_vi32, tor_neg_vi32_vi32);
#endif
*/
    rtst_passed();
}

void
test_abs(void)
{
    rtst_start("abs_vi32_vi32");
    rtst_log("testing nv.\n");
    abs_6(nv_abs_vi32_vi32);
/*
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    abs_6(tor_abs_vi32_vi32);
    random_test(nv_abs_vi32_vi32, tor_abs_vi32_vi32);
#endif
*/
    rtst_passed();
}

void
test_satneg(void)
{
    rtst_start("satneg_vi32_vi32");
    rtst_log("testing nv.\n");
    satneg_6(nv_satneg_vi32_vi32);
/*
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satneg_6(tor_satneg_vi32_vi32);
    random_test(nv_satneg_vi32_vi32, tor_satneg_vi32_vi32);
#endif
*/
    rtst_passed();
}

void
test_satabs(void)
{
    rtst_start("satabs_vi32_vi32");
    rtst_log("testing nv.\n");
    satabs_6(nv_satabs_vi32_vi32);
/*
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satabs_6(tor_satabs_vi32_vi32);
    random_test(nv_satabs_vi32_vi32, tor_satabs_vi32_vi32);
#endif
*/
    rtst_passed();
}

void
test_countlz(void)
{
    rtst_start("countlz_vi32_vi32");
    rtst_log("testing nv.\n");
    countlz_6(nv_countlz_vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    countlz_6(tor_countlz_vi32_vi32);
    random_test(nv_countlz_vi32_vi32, tor_countlz_vi32_vi32);
#endif
    rtst_passed();
}

void
test_swapb(void)
{
    rtst_start("swapb_vi32_vi32");
    rtst_log("testing nv.\n");
    swapb_explicit(nv_swapb_vi32_vi32);
/*
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    swapb_explicit(tor_swapb_vi32_vi32);
    random_test(nv_swapb_vi32_vi32, tor_swapb_vi32_vi32);
#endif
*/
    rtst_passed();
}


int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_copy();
    test_neg();
    test_abs();
    test_satneg();
    test_satabs();
    test_countlz();
    test_swapb();

    rtst_exit();

    return 0;
}
