/*
   $Header: /u/spert/src/intvec/testsuite/RCS/vecele_vi32i32_vi32_test.c,v 1.2 1996/01/03 21:15:07 davidj Exp $

   Test element wise vector routines of the form op_vfvf_vf.
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"


typedef void (*rtn_t)(size_t len, const intv_int32_t* a, intv_int32_t b,
		      intv_int32_t* res);

/* Performs one test given input values and expected output. */

void
onetest_normal(rtn_t rtn, size_t len,
	       const intv_int32_t* a, intv_int32_t b,
	       const intv_int32_t* correct)
{
    /* Create padded copies of inputs. */
    intv_int32_t* c_a = rtst_padvec_clone_vi32(len, a);
    intv_int32_t* c_res = rtst_padvec_new_vi32(len);

    /* Run routine. */
    rtn(len, c_a, b, c_res);

    /* Check result. */
    rtst_checkeq_vi32vi32(len, c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_vi32vi32(len, a, c_a);

    /* Free and check padding. */
    rtst_padvec_del_vi32(c_a);
    rtst_padvec_del_vi32(c_res);
}

void
add1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, 1, -1, 123, -456, 0x7fffffff };
    intv_int32_t b = 3;
    const intv_int32_t correct[] = { 3, 4, 2, 126, -453, 0x80000002 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
sub1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, 1, -1, 123, -456, 0x7fffffff };
    intv_int32_t b = -3;
    const intv_int32_t correct[] = { 3, 4, 2, 126, -453, 0x80000002 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
    
}

void
satadd1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, 1, -1, 123, -456, INTV_INT32_MAX };
    intv_int32_t b = 3;
    const intv_int32_t correct[] = { 3, 4, 2, 126, -453, INTV_INT32_MAX };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
satsub1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, 1, -1, 123, -456, INTV_INT32_MAX };
    intv_int32_t b = -3;
    const intv_int32_t correct[] = { 3, 4, 2, 126, -453, INTV_INT32_MAX };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
    
}

void
mul1(rtn_t rtn)
{
    const intv_int32_t a[] = { 0, 1, -1, 123, 0x7fffffff };
    intv_int32_t b = -3;
    const intv_int32_t correct[] = { 0, -3, 3, -369, -0x7ffffffd };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
    
}

/* Compare the results of two vector routines */
void
random_test(rtn_t nv_rtn, rtn_t test_rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    intv_int32_t *a = rtst_padvec_new_vi32(maxlen);
    intv_int32_t b;
    intv_int32_t *res = rtst_padvec_new_vi32(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_badrand_vi32(len, a);
        b = rtst_badrand_i32();
        nv_rtn(len, a, b, res);

        rtst_log("len=%d\n", len);

        onetest_normal(test_rtn, len, a, b, res);
    }

    rtst_padvec_del_vi32(a);
    rtst_padvec_del_vi32(res);
}

void
test_add(void)
{
    rtst_start("add_vi32i32_vi32");
    rtst_log("testing nv.\n");
    add1(nv_add_vi32i32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    add1(tor_add_vi32i32_vi32);
    random_test(nv_add_vi32i32_vi32, tor_add_vi32i32_vi32);
#endif
    rtst_passed();
}

void
test_sub(void)
{
    rtst_start("sub_vi32i32_vi32");
    rtst_log("testing nv.\n");
    sub1(nv_sub_vi32i32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    sub1(tor_sub_vi32i32_vi32);
    random_test(nv_sub_vi32i32_vi32, tor_sub_vi32i32_vi32);
#endif
    rtst_passed();
}

void
test_mul(void)
{
    rtst_start("mul_vi32i32_vi32");
    rtst_log("testing nv.\n");
    mul1(nv_mul_vi32i32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    mul1(tor_mul_vi32i32_vi32);
    random_test(nv_mul_vi32i32_vi32, tor_mul_vi32i32_vi32);
#endif
    rtst_passed();
}

void
test_satadd(void)
{
    rtst_start("satadd_vi32i32_vi32");
    rtst_log("testing nv.\n");
    satadd1(nv_satadd_vi32i32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satadd1(tor_satadd_vi32i32_vi32);
    random_test(nv_satadd_vi32i32_vi32, tor_satadd_vi32i32_vi32);
#endif
    rtst_passed();
}

void
test_satsub(void)
{
    rtst_start("satsub_vi32i32_vi32");
    rtst_log("testing nv.\n");
    satsub1(nv_satsub_vi32i32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satsub1(tor_satsub_vi32i32_vi32);
    random_test(nv_satsub_vi32i32_vi32, tor_satsub_vi32i32_vi32);
#endif
    rtst_passed();
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_add();
    test_sub();
    test_mul();

    test_satadd();
    test_satsub();

    rtst_exit();

    return 0;
}
