// $Header: /n/marmite/db/davidj/spert/src/quicknet/testsuite/RCS/SentDB_test.cc,v 1.3 1995/12/07 21:13:38 davidj Exp $
//
// Various tests for QN_SentDB classes
// davidj - Mon Dec  4 18:10:18 1995

#include <assert.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>

#include "QuickNet.h"
#include "rtst.h"
#include "fltvec.h"

void
SentDB_PFile_test(int debug, const char* pfile, long num_sent, long num_frames,
		  long num_ftrs, long num_labels, long num_targets)
{
    size_t i;

    rtst_start("SentDB_PFile");
    FILE* testfile = fopen(pfile, "r");
    assert(testfile!=NULL);

    QN_SentDB_PFile db(debug, testfile,
		       0, QN_SentDB_PFile::ALL_FEATURES,
		       0, QN_SentDB_PFile::ALL_LABELS,
		       0, 0);

    // Check the basic attributes of the PFile
    rtst_assert(db.num_sentences()==num_sent);
    rtst_assert(db.num_frames()==num_frames);
    rtst_assert(db.num_features_in_frame()==num_ftrs);
    rtst_assert(db.num_labels_in_frame()==num_labels);
    rtst_assert(db.num_targets_in_frame()==num_targets);

    // Sum all the number of frames in each sentence, and ensure the total
    // is the sum as the number of frames in the whole database
    {
	size_t acc_frames = 0;
	
	for (i=0; i<db.num_sentences(); i++)
	{
	    acc_frames += db.num_frames_in_sents(i, 1);
	}
	rtst_assert(acc_frames==num_frames);
    }

    // Test reading features
    {
	const int first_sent = 3;
	const int num_sents = 5;
	assert(db.num_sentences()>=first_sent+num_sents);
	size_t frames = db.num_frames_in_sents(first_sent, num_sents);
	assert(frames>num_sents*2);
	const size_t bufsize = frames * num_ftrs;

	float* buf = rtst_padvec_new_vf(bufsize);
	copy_f_vf(bufsize, -123456.7f, buf);

	size_t cnt = db.get_sentences(first_sent, num_sents,
				      NULL, NULL, buf, NULL, NULL);
	rtst_log("%lu frames in sentences %i to %i.\n",
		 (unsigned long) cnt, first_sent, first_sent+num_sents-1);
	rtst_assert(cnt==frames);
	rtst_assert(buf[(num_ftrs*frames)-1]!=-123456.7f);
	rtst_padvec_del_vf(buf);
    }

    // Test reading labels
    {
	const int first_sent = 5;
	const int num_sents = 7;
	assert(db.num_sentences()>=first_sent+num_sents);
	size_t frames = db.num_frames_in_sents(first_sent, num_sents);
	assert(frames>num_sents*2);
	const size_t bufsize = frames * num_labels;

	QNUInt32* lbuf = (QNUInt32*) rtst_padvec_new_vi32(bufsize);
	copy_i32_vi32(bufsize, 99999, (intv_int32_t*) lbuf);

	size_t cnt = db.get_sentences(first_sent, num_sents,
				      NULL, NULL, NULL, lbuf, NULL);
	rtst_log("%lu frames in sentences %i to %i.\n",
		 (unsigned long) cnt, first_sent, first_sent+num_sents-1);
	rtst_assert(cnt==frames);
	rtst_assert(lbuf[(num_labels*frames)-1]!=99999);
	rtst_padvec_del_vi32(lbuf);
    }

    // Test reading sentence and frame numbers
    {
	const int first_sent = 11;
	const int num_sents = 2;
	assert(db.num_sentences()>=first_sent+num_sents);
	size_t frames = db.num_frames_in_sents(first_sent, num_sents);
	assert(frames>num_sents*2);

	QNUInt32* sentnobuf = (QNUInt32*) rtst_padvec_new_vi32(frames);
	QNUInt32* framenobuf = (QNUInt32*) rtst_padvec_new_vi32(frames);
	size_t ncnt = db.get_sentences(first_sent, num_sents,
				       sentnobuf, framenobuf,
				       NULL, NULL, NULL);
	rtst_assert(ncnt==db.num_frames_in_sents(first_sent, num_sents));
	size_t frames_first = db.num_frames_in_sents(first_sent, 1);
	for (i=0; i<frames_first; i++)
	{
	    rtst_assert(sentnobuf[i]==first_sent);
	    rtst_assert(framenobuf[i]==i);
	}
	rtst_log("index data for %lu frames in sentence %lu correct.\n",
		 (unsigned long) frames_first,
		 (unsigned long) first_sent);
	size_t frames_second = db.num_frames_in_sents(first_sent+1, 1);
	for (i=0; i<frames_second; i++)
	{
	    rtst_assert(sentnobuf[frames_first+i]==first_sent+1);
	    rtst_assert(framenobuf[frames_first+i]==i);
	}
	rtst_log("index data for %lu frames in sentence %lu correct.\n",
		 (unsigned long) frames_second,
		 (unsigned long) first_sent+1);

	rtst_padvec_del_vi32((rtst_int32) sentnobuf);
	rtst_padvec_del_vi32((rtst_int32) framenobuf);
    }

    rtst_passed();
}

int
main(int argc, char* argv[])
{
    int arg;
    int debug = 0;		// Verbosity level

    arg = rtst_args(argc, argv);
    assert(arg==argc-6);
    QN_logger = new QN_Logger_Simple(rtst_logfile, stderr,
				     "SentDB_test");
    const char* pfile = argv[arg++];
    long num_sents = strtol(argv[arg++], NULL, 0);
    long num_frames = strtol(argv[arg++], NULL, 0);
    long num_ftrs = strtol(argv[arg++], NULL, 0);
    long num_labels = strtol(argv[arg++], NULL, 0);
    long num_targets = strtol(argv[arg++], NULL, 0);
    if (rtst_logfile!=NULL)
	debug = 99;
    SentDB_PFile_test(debug, pfile, num_sents, num_frames,
		      num_ftrs, num_labels, num_targets);
    rtst_exit();
}
